/******************************************************************************
 * Copyright (c) 1998 by Mindbright Technology AB, Stockholm, Sweden.
 *
 *                        All Rights Reserved
 *
 *               www.mindbright.se, info@mindbright.se
 *
 * This program and its source-code may be used FREE of charge for strictly
 * NON-COMMERCIAL purposes. Companies or persons wishing to make non-commercial
 * software that in whole or in part contains or is derived from this program
 * or any part thereof must inform the original author of such work. The
 * original author must also be given attribution as the author of used parts.
 * For COMMERCIAL use all parts of this program originally developed by
 * Mindbright Technology must be LICENSED under Mindbright Technology's
 * SHAREWARE license, please contact Mindbright Technology for information on
 * this. This copyright notice must allways be included and unchanged.
 *
 * IF LICENSED FREE OF CHARGE AND/OR IF NOT STATED OTHERWISE THIS PROGRAM COMES
 * WITH NO WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, IN NO EVENT SHALL
 * MINDBRIGHT BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR
 * ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER
 * IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
 * OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *****************************************************************************/
package mindbright.ssh;

import java.net.*;
import java.io.*;
import java.security.*;
import java.math.BigInteger;
import java.util.Vector;

import mindbright.security.*;
import mindbright.terminal.*;

public class SSHClient extends SSH implements Runnable {

  public static String copyright() {
    return "Copyright (c) 1998 by Mindbright Technology AB, Stockholm, Sweden";
  }

  protected InetAddress serverAddr;
  protected InetAddress localAddr;
  protected String      serverName;
  protected int         port;

  protected String username;
  protected String password;
  protected String srvVersion;

  boolean keepRunning;
  boolean doCommandShell;

  // Local port forwarding
  //
  static public class LocalForward {
    protected int    localPort;
    protected String remoteHost;
    protected int    remotePort;
    protected String plugin;
    public LocalForward(int localPort, String remoteHost, int remotePort, String plugin) {
      this.localPort  = localPort;
      this.remoteHost = remoteHost;
      this.remotePort = remotePort;
      this.plugin     = plugin;
    }
  }

  // Remote port forwarding
  //
  static public class RemoteForward {
    protected int    remotePort;
    protected String localHost;
    protected int    localPort;
    protected String plugin;
    public RemoteForward(int remotePort, String localHost, int localPort, String plugin) {
      this.remotePort = remotePort;
      this.localHost  = localHost;
      this.localPort  = localPort;
      this.plugin     = plugin;
    }
  }
    
  protected Vector localForwards;
  protected Vector remoteForwards;

  protected SSHChannelController controller;
  protected SSHStdIO             console;

  protected Socket               sshSocket;
  protected BufferedInputStream  sshIn;
  protected BufferedOutputStream sshOut;

  protected Cipher sndCipher;
  protected Cipher rcvCipher;

  public SSHClient() {
    port                = DEFAULTPORT;
    this.serverName     = "localhost";
    this.localForwards  = new Vector();
    this.remoteForwards = new Vector();
  }

  public void setConsole(SSHStdIO console) {
    this.console        = console;
  }

  public String getServerName() {
    return serverName;
  }

  public InetAddress getServerAddr() {
    return serverAddr;
  }

  public InetAddress getLocalAddr() {
    return localAddr;
  }

  public String getServerVersion() {
    return srvVersion;
  }

  public int getPort() {
    return port;
  }

  public String getUsername() {
    return username;
  }

  public void setServerName(String server) {
    int i;
    if((i = server.indexOf(':')) != -1) {
      this.port = Integer.parseInt(server.substring(i));
      this.serverName = server.substring(0, i);
    } else
      this.serverName = server;
  }

  public void setPort(int port) {
    this.port = port;
  }

  public void setUsername(String username) {
    this.username = username;
  }

  public void setPassword(String password) {
    this.password = password;
  }

  public void addLocalPortForward(String fwdSpec) throws Exception {
    int    localPort;
    String remoteHost;
    int    remotePort;
    int    d1, d2, pd;
    String tmp, plugin;

    if(fwdSpec.charAt(0) == '/') {
      int i = fwdSpec.lastIndexOf('/');
      if(i == 0)
	throw new Exception("Invalid port forward spec.");
      plugin = fwdSpec.substring(1, i);
      fwdSpec = fwdSpec.substring(i + 1);
    } else
      plugin = "default";

    d1 = fwdSpec.indexOf(':');
    d2 = fwdSpec.lastIndexOf(':');
    if(d1 == d2)
      throw new Exception("Invalid port forward spec.");

    tmp        = fwdSpec.substring(0, d1);
    localPort  = Integer.parseInt(tmp);
    remoteHost = fwdSpec.substring(d1 + 1, d2);
    tmp        = fwdSpec.substring(d2 + 1);
    remotePort = Integer.parseInt(tmp);
    addLocalPortForward(localPort, remoteHost, remotePort, plugin);
  }

  public void addLocalPortForward(int localPort, String remoteHost, int remotePort, String plugin) {
    localForwards.addElement(new LocalForward(localPort, remoteHost, remotePort, plugin));
  }

  public void delLocalPortForward(int port) {
    for(int i = 0; i < localForwards.size(); i++) {
      LocalForward fwd = (LocalForward) localForwards.elementAt(i);
      if(fwd.localPort == port) {
	localForwards.removeElementAt(i);
	break;
      }
    }
  }

  public void addRemotePortForward(String fwdSpec) throws Exception {
    int    remotePort;
    int    localPort;
    String localHost;
    int    d1, d2;
    String tmp, plugin;

    if(fwdSpec.charAt(0) == '/') {
      int i = fwdSpec.lastIndexOf('/');
      if(i == 0)
	throw new Exception("Invalid port forward spec.");
      plugin = fwdSpec.substring(1, i);
      fwdSpec = fwdSpec.substring(i + 1);
    } else
      plugin = "default";

    d1 = fwdSpec.indexOf(':');
    d2 = fwdSpec.lastIndexOf(':');
    if(d1 == d2)
      throw new Exception("Invalid port forward spec.");

    tmp        = fwdSpec.substring(0, d1);
    remotePort = Integer.parseInt(tmp);
    localHost  = fwdSpec.substring(d1 + 1, d2);
    tmp        = fwdSpec.substring(d2 + 1);
    localPort  = Integer.parseInt(tmp);
    addRemotePortForward(remotePort, localHost, localPort, plugin);
  }

  public void addRemotePortForward(int remotePort, String localHost, int localPort, String plugin) {
    remoteForwards.addElement(new RemoteForward(remotePort, localHost, localPort, plugin));
  }

  public void delRemotePortForward(int port) {
    for(int i = 0; i < remoteForwards.size(); i++) {
      RemoteForward fwd = (RemoteForward) remoteForwards.elementAt(i);
      if(fwd.remotePort == port) {
	remoteForwards.removeElementAt(i);
	break;
      }
    }
  }

  protected void print(String str) {
    if(console != null)
      console.print(str);
  }
  protected void println(String str) {
    if(console != null)
      console.println(str);
  }
  protected void updateTitle() {
    if(console != null)
      console.updateTitle();
  }

  public void run() {
    println(copyright());
    println("");

    print("Initializing please wait...");
    // We do this here to seed the algorithm which takes some time...
    SecureRandom dummy = new SecureRandom();
    println("done");

    keepRunning    = true;
    doCommandShell = false;
    updateTitle();

    // !!!
    initiatePlugins();

    while(keepRunning) {
      try {
	println("(press ^D before logging in to enter local command-shell)");
	if(console != null && console.isDumb())
	  println("(...you might have to press ENTER also...");

	println("");

	if(console != null)
	  setServerName(console.promptLine("SSH-server: ", serverName, false));

	serverAddr = InetAddress.getByName(serverName);
	sshSocket  = new Socket(serverName, port);
	sshIn      = new BufferedInputStream(sshSocket.getInputStream(), 8192);
	sshOut     = new BufferedOutputStream(sshSocket.getOutputStream());
	localAddr  = sshSocket.getLocalAddress();

	updateTitle();

	negotiateVersion();

	println("Connected to server running " + srvVersion);
	println("");
	updateTitle();

	receiveServerData();

	// !!! Hardcoded use of Blowfish and only supports password-authentication !!!
	if(!isCipherSupported(CIPHER_BLOWFISH))
	  throw new Exception("Sorry, we only support Blowfish cipher");
	if(!isAuthTypeSupported(AUTH_PASSWORD))
	  throw new Exception("Sorry, we only supoort password authentication");

	generateSessionId();
	generateSessionKey();

	sndCipher = new BlowfishCipher();
	rcvCipher = new BlowfishCipher();
	sndCipher.setKey(sessionKey);
	rcvCipher.setKey(sessionKey);
	sendSessionKey(CIPHER_BLOWFISH);

	// !!!
	// At this stage the communication is encrypted
	// !!!

	if(console != null)
	  setUsername(console.promptLine(serverName + " login: ", username, false));

	updateTitle();

	authenticateUser();

	controller = new SSHChannelController(this, sshIn, sshOut, sndCipher, rcvCipher, console);

	initiateSession();
	startService();

      } catch(SSHStdIO.CtrlDPressedException e) {
	doCommandShell = true;
      } catch(UnknownHostException e) {
	println("Unknown host: " + serverName);
      } catch(Exception e) {
	println("Error connecting to " + serverName + ", reason:");
	println(e.getMessage());
      }
      println("");
      srvVersion = null;
      updateTitle();

      if(console != null && doCommandShell) {
	println("");
	println("...entering local command-shell (type 'h' for help).");
	println("");
	keepRunning = console.doCommandShell();
	doCommandShell = false;
      }
    }
  }

  void negotiateVersion() throws IOException {
    byte[] buf = new byte[256];
    int    len;
    String verStr;

    len = sshIn.read(buf);

    // !!! For now...
    srvVersion = new String(buf, 0, len);
    if(srvVersion.indexOf("SSH-1.5") != 0 && srvVersion.indexOf("SSH-1.99") != 0)
      throw new IOException("Version mismatch, can't talk to server");

    // Strip trailing newline
    srvVersion = srvVersion.substring(0, srvVersion.length() - 1);

    verStr = getVersionId();
    verStr += "\n";
    buf    = verStr.getBytes();

    sshOut.write(buf);
    sshOut.flush();
  }

  void receiveServerData() throws IOException {
    SSHPduInputStream pdu = new SSHPduInputStream(SMSG_PUBLIC_KEY, null);
    pdu.readFrom(sshIn);
    int bits;
    BigInteger e, n;

    srvCookie = new byte[8];
    pdu.read(srvCookie, 0, 8);

    bits = pdu.readInt();
    e = pdu.readBigInteger();
    n = pdu.readBigInteger();
    srvServerKey = new KeyPair(new RSAPublicKey(e, n), null);

    bits = pdu.readInt();
    e = pdu.readBigInteger();
    n = pdu.readBigInteger();
    srvHostKey = new KeyPair(new RSAPublicKey(e, n), null);

    protocolFlags      = pdu.readInt();
    supportedCiphers   = pdu.readInt();
    supportedAuthTypes = pdu.readInt();
  }

  void generateSessionId() throws IOException {
    MessageDigest md5;
    byte[]        message;
    byte[]        srvKey = ((RSAPublicKey)srvServerKey.getPublic()).getN().toByteArray();
    byte[]        hstKey = ((RSAPublicKey)srvHostKey.getPublic()).getN().toByteArray();
    int           i, len = srvKey.length + hstKey.length + srvCookie.length;

    if(srvKey[0] == 0)
      len -= 1;
    if(hstKey[0] == 0)
      len -= 1;

    message = new byte[len];

    if(hstKey[0] == 0) {
      System.arraycopy(hstKey, 1, message, 0, hstKey.length - 1);
      len = hstKey.length - 1;
    } else {
      System.arraycopy(hstKey, 0, message, 0, hstKey.length);
      len = hstKey.length;
    }

    if(srvKey[0] == 0) {
      System.arraycopy(srvKey, 1, message, len, srvKey.length - 1);
      len += srvKey.length - 1;
    } else {
      System.arraycopy(srvKey, 0, message, len, srvKey.length);
      len += srvKey.length;
    }

    System.arraycopy(srvCookie, 0, message, len, srvCookie.length);

    try {
      md5 = MessageDigest.getInstance("MD5");
      sessionId = md5.digest(message);
    } catch(Exception e) {
      throw new IOException("MD5 not implemented, can't generate session-id");
    }
  }

  void generateSessionKey() {
    SecureRandom rand = new SecureRandom();
    sessionKey = new byte[SESSION_KEY_LENGTH / 8];
    rand.nextBytes(sessionKey);
  }

  void sendSessionKey(int cipherType) throws IOException {
    byte[]             key = new byte[sessionKey.length + 1];
    BigInteger         encKey;
    RSACipher          rsa;
    SSHPduOutputStream pdu;

    key[0] = 0;
    System.arraycopy(sessionKey, 0, key, 1, sessionKey.length);

    for(int i = 0; i < sessionId.length; i++)
      key[i + 1] ^= sessionId[i];

    encKey = new BigInteger(key);

    if(((RSAPublicKey)(srvServerKey.getPublic())).bitLength() <
       ((RSAPublicKey)(srvHostKey.getPublic())).bitLength()) {
      BigInteger padded;
      rsa = new RSACipher(srvServerKey);
      padded = rsaPad(encKey, ((RSAPublicKey)srvServerKey.getPublic()).bitLength());
      encKey = rsa.doPublic(padded);
      rsa = new RSACipher(srvHostKey);
      padded = rsaPad(encKey, ((RSAPublicKey)srvHostKey.getPublic()).bitLength());
      encKey = rsa.doPublic(padded);
    } else {
      BigInteger padded;
      rsa = new RSACipher(srvHostKey);
      padded = rsaPad(encKey, ((RSAPublicKey)srvHostKey.getPublic()).bitLength());
      encKey = rsa.doPublic(padded);
      rsa = new RSACipher(srvServerKey);
      padded = rsaPad(encKey, ((RSAPublicKey)srvServerKey.getPublic()).bitLength());
      encKey = rsa.doPublic(padded);
    }

    pdu = new SSHPduOutputStream(CMSG_SESSION_KEY, null);
    pdu.writeByte((byte)cipherType);
    pdu.write(srvCookie, 0, srvCookie.length);
    pdu.writeBigInteger(encKey);
    // !!! TODO: check this pdu.writeInt(PROTOFLAG_SCREEN_NUMBER | PROTOFLAG_HOST_IN_FWD_OPEN);
    pdu.writeInt(protocolFlags);
    pdu.writeTo(sshOut);

    expectSuccess();
  }

  void authenticateUser() throws IOException {

    SSHPduOutputStream outpdu;
    SSHPduInputStream  inpdu;

    outpdu = new SSHPduOutputStream(CMSG_USER, sndCipher);
    outpdu.writeString(username);
    outpdu.writeTo(sshOut);

    inpdu = new SSHPduInputStream(MSG_ANY, rcvCipher);
    inpdu.readFrom(sshIn);

    if(inpdu.type == SMSG_SUCCESS) {
      println("Authenticated directly by server, no password required");
      return;
    } else if(inpdu.type != SMSG_FAILURE) {
      throw new IOException("Protocol error: got " + inpdu.type + 
			    " in response to CMSG_USER");
    }

    if(console != null)
      setPassword(console.promptLine(username + "@" + serverName + "'s password: ", "", true));

    outpdu = new SSHPduOutputStream(CMSG_AUTH_PASSWORD, sndCipher);
    outpdu.writeString(password);
    outpdu.writeTo(sshOut);

    inpdu = new SSHPduInputStream(MSG_ANY, rcvCipher);
    inpdu.readFrom(sshIn);

    if(inpdu.type == SMSG_SUCCESS) {
      return;
    } else if(inpdu.type != SMSG_FAILURE)
      throw new IOException("Protocol error: got " + inpdu.type + " in response to SSH_CMSG_USER");
    else
      throw new IOException("Permission denied");
  }

  void initiateSession() throws IOException {
    // !!! TODO: Handle command and more options
    //
    requestPTY();
    initiateTunnels();
    requestShell();
    // !!!
    // At this stage we can't send more options/forward-requests
    // the server has now entered it's service-loop.
  }

  void initiatePlugins() {
    SSHProtocolPlugin.addPlugin("default", new SSHProtocolPlugin());

    // !!! TODO: Handle plugins generically through per user properties
    // should only be done once at startup (or when a plugin is added).
    //
    SSHFtpPlugin ftpPlugin = new SSHFtpPlugin();
    SSHProtocolPlugin.addPlugin("ftp", ftpPlugin);
    ftpPlugin.initiate(this);
  }

  void initiateTunnels() throws IOException {
    int i;
    for(i = 0; i < localForwards.size(); i++) {
      LocalForward fwd = (LocalForward) localForwards.elementAt(i);
      requestLocalPortForward(fwd.localPort, fwd.remoteHost, fwd.remotePort, fwd.plugin);
    }
    for(i = 0; i < remoteForwards.size(); i++) {
      RemoteForward fwd = (RemoteForward) remoteForwards.elementAt(i);
      requestRemotePortForward(fwd.remotePort, fwd.localHost, fwd.localPort, fwd.plugin);
    }
  }

  void startService() throws Exception {
    if(console != null)
      console.serverConnect(controller, sndCipher);
    controller.start();

    if(console != null && console.isDumb())
      console.startDumbConsole();

    controller.waitForExit();

    if(console != null && console.isDumb())
      console.stopDumbConsole();
  }

  void requestPTY() throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_REQUEST_PTY, sndCipher);
    Terminal myTerminal = null;
    if(console != null)
      myTerminal = console.getTerminal();
    if(myTerminal != null) {
      outpdu.writeString(myTerminal.terminalType());
      outpdu.writeInt(myTerminal.rows());
      outpdu.writeInt(myTerminal.cols());
      outpdu.writeInt(myTerminal.vpixels());
      outpdu.writeInt(myTerminal.hpixels());
    } else {
      outpdu.writeString("");
      outpdu.writeInt(0);
      outpdu.writeInt(0);
      outpdu.writeInt(0);
      outpdu.writeInt(0);
    }
    outpdu.writeByte((byte)TTY_OP_END);
    outpdu.writeTo(sshOut);

    expectSuccess();
  }

  void requestLocalPortForward(int localPort, String remoteHost, int remotePort, String plugin) {
    controller.newListenChannel(localPort, remoteHost, remotePort, plugin);
  }

  void requestRemotePortForward(int remotePort, String localHost, int localPort, String plugin)
    throws IOException {

    SSHProtocolPlugin.getPlugin(plugin).remoteListener(remotePort, localHost, localPort,
						       controller);

    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_PORT_FORWARD_REQUEST, sndCipher);
    outpdu.writeInt(remotePort);
    outpdu.writeString(localHost);
    outpdu.writeInt(localPort);
    outpdu.writeTo(sshOut);
    expectSuccess();
  }

  void expectSuccess() throws IOException {
    SSHPduInputStream inpdu = null;
    try {
      inpdu = new SSHPduInputStream(SMSG_SUCCESS, rcvCipher);
      inpdu.readFrom(sshIn);
    } catch (IOException e) {
      if(inpdu != null && inpdu.type == MSG_DISCONNECT)
	throw new IOException("Server disconnected: " + inpdu.readString());
      else
	throw e;
    }
  }

  void requestCommand(String command) throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_EXEC_CMD, null);
    outpdu.writeString(command);
    outpdu.writeTo(sshOut);
  }

  void requestShell() throws IOException {
    SSHPduOutputStream outpdu = new SSHPduOutputStream(CMSG_EXEC_SHELL, sndCipher);
    outpdu.writeTo(sshOut);
  }

  void setInteractive() {
    try {
      sshSocket.setTcpNoDelay(true);
    } catch (SocketException e) {
      println("Error setting interactive mode: " + e.getMessage());
    }
  }

  public String getVersionId() {
    return "SSH-" + VER_MAJOR + "." + VER_MINOR + "-" + VER_SSH;
  }

}
